from collections import OrderedDict
from enum import Enum

from typing import Type, List

from pytconf.extended_enum import str_to_enum_value, ExtendedEnum


class EnumSubset:

    def __init__(self, enum_type: Type[Enum], list_of_values: List[Type[Enum]]) -> None:
        self.enum_type = enum_type
        # TODO: this should actually be an ordered set
        self.selected = OrderedDict()
        for value in list_of_values:
            self.add(value)

    def add(self, enum_value):
        assert (
            enum_value in enum_value.__class__.__members__.values()
        ), "bad value {}".format(enum_value)
        self.selected[enum_value] = None

    def delete(self, enum_value):
        del self.selected[enum_value]

    def yield_values(self):
        for value in self.selected.keys():
            yield value

    def values(self):
        return self.selected.keys()

    def __contains__(self, item):
        return item in self.selected

    def has_value(self, item):
        return item in {x.value for x in self.selected}

    def list_of_strings(self):
        my_list = []
        for x in self.selected:
            my_list.append(x.name)
        return my_list

    def to_string(self):
        return ",".join(self.list_of_strings())

    @classmethod
    def from_enum_all(cls, e):
        return g_from_enum_all(e)

    @classmethod
    def from_string(cls, e, s):
        return g_from_string(e, s)


def g_from_enum_all(e: Type[ExtendedEnum]) -> EnumSubset:
    return EnumSubset(enum_type=e, list_of_values=e.get_list_of_all_values(),)


def g_from_string(e: Type[Enum], s: str) -> EnumSubset:
    r = EnumSubset(enum_type=e, list_of_values=[])
    for name in s.split(","):
        v = str_to_enum_value(s=name, e=e)
        r.add(v)
    return r


def enum_subset_to_str(e: EnumSubset) -> str:
    return e.to_string()



from typing import List, Union


def convert_str_to_int_or_none(s: str) -> Union[int, None]:
    if s == "None":
        return None
    return int(s)


def convert_int_or_none_to_str(v: Union[int, None]) -> str:
    if v is None:
        return "None"
    return str(v)


def convert_str_to_int(s: str) -> int:
    return int(s)


def convert_str_to_int_default(i: int, s: str) -> int:
    if s.startswith("+"):
        return i + int(s[1:])
    if s.startswith("-"):
        return i - int(s[1:])
    if s.startswith("*"):
        return i * int(s[1:])
    if s.startswith("/"):
        return i // int(s[1:])
    assert False, "you should not get here"


def convert_int_to_str(i: int) -> str:
    return str(i)


def convert_str_to_bool(s: str) -> bool:
    if s in {"TRUE", "true", "True", "yes", "Yes", "1", "y", "Y", "t", "T"}:
        return True
    return False


def convert_bool_to_str(b: bool) -> str:
    return str(b)


def convert_str_to_list_int(s: str) -> List[int]:
    return [int(x) for x in s.split(",")]


def convert_list_int_to_str(li: List[int]) -> str:
    return ",".join([str(x) for x in li])


def convert_str_to_list_str(s: str) -> List[str]:
    return s.split(",")


def convert_list_str_to_str(li: List[str]) -> str:
    return ",".join(li)


def convert_str_to_str(e: str) -> str:
    return e


def convert_str_to_str_or_none(s: str) -> Union[str, None]:
    if s == "None":
        return None
    return s


def convert_str_or_none_to_str(v: Union[str, None]) -> str:
    if v is None:
        return "None"
    return v



from pytconf.param import ParamCreator # noqa F401
from pytconf.config import ( # noqa F401
    Config,
    register_main,
    register_function,
    register_function_group,
    config_arg_parse_and_launch,
    register_endpoint,
    get_free_args,
    write_config_file_json_user,
    write_config_file_json_system,
    rm_config_file_json_user,
    rm_config_file_json_system,
)



class ParamCollector:
    def __init__(self):
        self.data = []

    def add_data(self, item):
        self.data.append(item)

    def clear(self):
        self.data = []

    def get_item(self, i):
        return self.data[i]

    def yield_items(self):
        for item in self.data:
            yield item


the_collector: ParamCollector = ParamCollector()



import abc
from enum import Enum
from typing import List, Union, Callable, Any, Type

from pyfakeuse import fake_use

from pytconf.enum_subset import EnumSubset

from pytconf.extended_enum import str_to_enum_value, enum_type_to_list_str

from pytconf.convert import convert_str_to_str, convert_str_to_int, convert_int_to_str, convert_str_to_int_default, \
    convert_str_to_list_int, convert_list_int_to_str, convert_str_to_list_str, convert_list_str_to_str, \
    convert_str_to_int_or_none, convert_int_or_none_to_str, convert_str_to_str_or_none, convert_str_or_none_to_str, \
    convert_str_to_bool, convert_bool_to_str
from pytconf.param_collector import the_collector

NO_HELP = "No help for this configuration option"


class Unique:
    pass


NO_DEFAULT = Unique()


class Param(abc.ABC):
    """
    Parent class of all parameters of configuration
    """

    def __init__(
        self,
        help_string=NO_HELP,
        default: Any = NO_DEFAULT,
        type_name: str = None,
    ):
        super().__init__()
        self.help_string = help_string
        self.required = default is NO_DEFAULT
        self.default = default
        self.type_name = type_name

    def get_type_name(self):
        return self.type_name

    def collect(self):
        the_collector.add_data(self)

    @abc.abstractmethod
    def s2t(self, s: str) -> object:
        pass

    def s2t_generate_from_default(self, s: str) -> object:
        raise ValueError("we do not support generation from default")

    @abc.abstractmethod
    def t2s(self, t: object) -> str:
        pass

    def more_help(self) -> Union[str, None]:
        return fake_use(self)


class ParamFunctions(Param):
    """
    Parent class of all parameters of configuration
    """

    def __init__(
        self,
        help_string=NO_HELP,
        default=None,
        type_name=None,
        function_s2t: Callable = None,
        function_s2t_generate_from_default: Callable = None,
        function_t2s: Callable = None,
    ):
        super().__init__(
            help_string=help_string,
            default=default,
            type_name=type_name,
        )
        self.function_s2t = function_s2t
        self.function_t2s = function_t2s
        self.function_s2t_generate_from_default = function_s2t_generate_from_default
        super().collect()

    def s2t(self, s: str) -> Any:
        return self.function_s2t(s)

    def s2t_generate_from_default(self, s: str) -> Any:
        return self.function_s2t_generate_from_default(self.default, s)

    def t2s(self, t: Any) -> str:
        return self.function_t2s(t)


class ParamFilename(Param):
    def __init__(
        self,
        help_string: str = NO_HELP,
        default: str = NO_DEFAULT,
        type_name=None,
        suffixes: List[str] = None,
    ):
        super().__init__(
            help_string=help_string,
            default=default,
            type_name=type_name,
        )
        self.suffixes = suffixes
        super().collect()

    def s2t(self, s):
        if self.suffixes is not None:
            assert any(
                s.endswith(x) for x in self.suffixes
            ), "filename suffix is not accepted"
        return s

    def t2s(self, t):
        return t

    def more_help(self):
        if self.suffixes is None:
            return "no limitation on suffixes"
        return "allowed suffixes are {}".format(self.suffixes)


class ParamEnum(Param):
    def __init__(
        self,
        help_string=NO_HELP,
        default: Any = NO_DEFAULT,
        enum_type: Type[Enum] = None,
    ):
        super().__init__(
            help_string=help_string,
            default=default,
            type_name="enum",
        )
        self.enum_type = enum_type
        super().collect()

    def get_type_name(self):
        return "Enum[{}]".format(self.enum_type.__name__)

    def s2t(self, s: str) -> Any:
        return str_to_enum_value(s, self.enum_type)

    def t2s(self, t: Any) -> str:
        return t.name

    def more_help(self):
        return "allowed values {}".format(enum_type_to_list_str(self.enum_type))


class ParamEnumSubset(Param):
    def __init__(
        self,
        help_string=NO_HELP,
        default: Any = NO_DEFAULT,
        enum_type: Type[Enum] = None,
    ):
        super().__init__(
            help_string=help_string,
            default=default,
            type_name="enum",
        )
        self.enum_type = enum_type
        super().collect()

    def get_type_name(self):
        return "EnumSubset[{}]".format(self.enum_type.__name__)

    def s2t(self, s: str) -> EnumSubset:
        return EnumSubset.from_string(e=self.enum_type, s=s)

    def t2s(self, t: EnumSubset) -> str:
        return t.to_string()

    def s2t_generate_from_default(self, s: str) -> EnumSubset:
        pass

    def more_help(self):
        return "allowed values {}".format(enum_type_to_list_str(self.enum_type))


class ParamChoice(Param):
    def __init__(
        self,
        help_string=NO_HELP,
        default: str = NO_DEFAULT,
        choice_list: List[str] = None
    ):
        super().__init__(
            help_string=help_string,
            default=default,
            type_name="Choice",
        )
        self.choice_list = choice_list
        super().collect()

    def s2t(self, s: str) -> Any:
        return s

    def t2s(self, t: Any) -> str:
        return t

    def more_help(self):
        return "allowed values {}".format(",".join(self.choice_list))


class ParamCreator:
    """
    Static namespace with all the param creation functions.
    """

    @staticmethod
    def create_int(
        help_string: str = NO_HELP,
        default: int = NO_DEFAULT,
    ) -> int:
        """
        Create an int parameter
        :param help_string:
        :param default:
        :return:
        """
        ParamFunctions(
            help_string=help_string,
            default=default,
            type_name="int",
            function_s2t=convert_str_to_int,
            function_t2s=convert_int_to_str,
            function_s2t_generate_from_default=convert_str_to_int_default,
        )
        if default is NO_DEFAULT:
            return 0
        return default

    @staticmethod
    def create_list_int(
        help_string: str = NO_HELP,
        # This is because of pylint
        # default: List[int] = NO_DEFAULT,
        # pylint: disable=dangerous-default-value
        default: List[int] = [],
    ) -> List[int]:
        """
        Create a List[int] parameter
        :param help_string:
        :param default:
        :return:
        """
        ParamFunctions(
            help_string=help_string,
            default=default,
            type_name="List[int]",
            function_s2t=convert_str_to_list_int,
            function_t2s=convert_list_int_to_str,
        )
        if default is NO_DEFAULT:
            return []
        return default

    @staticmethod
    def create_list_str(
        help_string: str = NO_HELP,
        # This is because of pylint
        # default: List[str] = NO_DEFAULT,
        # pylint: disable=dangerous-default-value
        default: List[str] = [],
    ) -> List[str]:
        """
        Create a List[str] parameter
        :param help_string:
        :param default:
        :return:
        """
        ParamFunctions(
            help_string=help_string,
            default=default,
            type_name="List[str]",
            function_s2t=convert_str_to_list_str,
            function_t2s=convert_list_str_to_str,
        )
        if default is NO_DEFAULT:
            return []
        return default

    @staticmethod
    def create_int_or_none(
        help_string: str = NO_HELP,
        default: Union[int, None] = NO_DEFAULT,
    ) -> Union[int, None]:
        """
        Create an int parameter
        :param help_string:
        :param default:
        :return:
        """
        ParamFunctions(
            help_string=help_string,
            default=default,
            type_name="Union[int, None]",
            function_s2t=convert_str_to_int_or_none,
            function_t2s=convert_int_or_none_to_str,
        )
        if default is NO_DEFAULT:
            return None
        return default

    @staticmethod
    def create_str(
        help_string: str = NO_HELP,
        default: str = NO_DEFAULT,
    ) -> str:
        """
        Create a string parameter
        :param help_string:
        :param default:
        :return:
        """
        ParamFunctions(
            help_string=help_string,
            default=default,
            type_name="str",
            function_s2t=convert_str_to_str,
            function_t2s=convert_str_to_str,
        )
        return default

    @staticmethod
    def create_str_or_none(
        help_string: str = NO_HELP,
        default: Union[str, None] = NO_DEFAULT,
    ) -> Union[str, None]:
        """
        Create a string parameter
        :param help_string:
        :param default:
        :return:
        """
        ParamFunctions(
            help_string=help_string,
            default=default,
            type_name="str_or_none",
            function_s2t=convert_str_to_str_or_none,
            function_t2s=convert_str_or_none_to_str,
        )
        if default == NO_DEFAULT:
            return None
        return default

    @staticmethod
    def create_bool(
        help_string: str = NO_HELP,
        default: bool = NO_DEFAULT,
    ) -> bool:
        """
        Create a bool parameter
        :param help_string:
        :param default:
        :return:
        """
        ParamFunctions(
            help_string=help_string,
            default=default,
            type_name="bool",
            function_s2t=convert_str_to_bool,
            function_t2s=convert_bool_to_str,
        )
        if default is NO_DEFAULT:
            return False
        return default

    @staticmethod
    def create_new_file(
        help_string: str = NO_HELP,
        default: str = NO_DEFAULT,
        suffixes: List[str] = None,
    ) -> str:
        """
        Create a new file parameter
        :param help_string:
        :param default:
        :param suffixes:
        :return:
        """
        ParamFilename(
            help_string=help_string,
            default=default,
            type_name="new_file",
            suffixes=suffixes,
        )
        if default is NO_DEFAULT:
            return ""
        return default

    @staticmethod
    def create_existing_file(
        help_string: str = NO_HELP,
        default: str = NO_DEFAULT,
        suffixes: Union[List[str], None] = None,
    ) -> str:
        """
        Create a new file parameter
        :param help_string:
        :param default:
        :param suffixes:
        :return:
        """
        ParamFilename(
            help_string=help_string,
            default=default,
            type_name="existing_file",
            suffixes=suffixes,
        )
        if default is NO_DEFAULT:
            return ""
        return default

    @staticmethod
    def create_existing_folder(
        help_string: str = NO_HELP,
        default: str = NO_DEFAULT,
        suffixes: Union[List[str], None] = None,
    ) -> str:
        """
        Create a new folder parameter
        :param help_string:
        :param default:
        :param suffixes:
        :return:
        """
        ParamFilename(
            help_string=help_string,
            default=default,
            type_name="existing_folder",
            suffixes=suffixes,
        )
        if default is NO_DEFAULT:
            return ""
        return default

    @staticmethod
    def create_choice(
        choice_list: List[str],
        help_string: str = NO_HELP,
        default: str = NO_DEFAULT,
    ) -> str:
        """
        Create a choice config
        :param choice_list:
        :param help_string:
        :param default:
        :return:
        """
        ParamChoice(
            help_string=help_string,
            default=default,
            choice_list=choice_list,
        )
        if default is NO_DEFAULT:
            return ""
        return default

    @staticmethod
    def create_enum(
        enum_type: Type[Enum],
        help_string: str = NO_HELP,
        default: Type[Enum] = NO_DEFAULT,
    ) -> Type[Enum]:
        """
        Create an enum config
        :param enum_type:
        :param help_string:
        :param default:
        :return:
        """
        ParamEnum(
            help_string=help_string,
            default=default,
            enum_type=enum_type,
        )
        if default is NO_DEFAULT:
            return enum_type
        return default

    @staticmethod
    def create_enum_subset(
        enum_type: Type[Enum],
        help_string: str = NO_HELP,
        default: EnumSubset = NO_DEFAULT,
    ) -> EnumSubset:
        """
        Create an enum config
        :param enum_type:
        :param help_string:
        :param default:
        :return:
        """
        ParamEnumSubset(
            help_string=help_string,
            default=default,
            enum_type=enum_type,
        )
        if default is NO_DEFAULT:
            return EnumSubset(enum_type=enum_type, list_of_values=[])
        return default

    @staticmethod
    def create_existing_bucket(
        help_string: str = NO_HELP,
        default: str = NO_DEFAULT,
    ) -> str:
        """
        Create a bucket name on gcp
        :param help_string:
        :param default:
        :return:
        """
        ParamFunctions(
            help_string=help_string,
            default=default,
            type_name="bucket_name",
            function_s2t=convert_str_to_str,
            function_t2s=convert_str_to_str,
        )
        if default is NO_DEFAULT:
            return ""
        return default



class Registry:
    def __init__(self):
        self.names = set()
        self.configs = set()
        self.name_to_data = {}
        self.name_to_config = {}
        self.config_names = {}

    def register(self, config, name, data):
        if name in self.names:
            raise ValueError(
                f"pytconf: name [{name}] appears more than once"
            )
        self.names.add(name)
        self.configs.add(config)
        if config not in self.config_names:
            self.config_names[config] = {}
        self.name_to_data[name] = data
        self.name_to_config[name] = config
        self.config_names[config][name] = data

    def yield_configs(self):
        for config in self.configs:
            yield config

    def yield_names_for_config(self, config):
        yield from self.config_names[config].keys()

    def yield_name_data_for_config(self, config):
        yield from self.config_names[config].items()

    def get_data_for_name(self, name):
        return self.name_to_data[name]

    def get_config_for_name(self, name):
        return self.name_to_config[name]

    def has_name(self, name):
        return name in self.names


the_registry: Registry = Registry()



from enum import Enum
from typing import Type, List


class ExtendedEnum(Enum):
    @classmethod
    def get_list_of_all_values(cls):
        return cls.__members__.values()

    @classmethod
    def from_string(cls, s: str) -> Type[Enum]:
        for x in cls:
            if s == x.name:
                return x
        raise ValueError("cannot find value [{}]".format(s))


def enum_type_to_list_str(e: Type[Enum]) -> List[str]:
    a = [x.name for x in e]
    return a


def enum_value_to_str(ev: Type[Enum]) -> str:
    return ev.name


def str_to_enum_value(s, e: Type[Enum]) -> Type[Enum]:
    for x in e:
        if s == x.name:
            return x
    raise ValueError("cannot find value [{}]".format(s))


def str_to_enum_int(s: str, e: Type[Enum]) -> int:
    for x in e:
        if s == x.name:
            return x.value
    raise ValueError("cannot find value [{}]".format(s))


def enum_int_to_name(i: int, e: Type[Enum]) -> str:
    for x in e:
        if i == x.value:
            return x.name
    raise ValueError("cannot find value [{}]".format(i))



from termcolor import colored, cprint


def print_highlight(text):
    cprint(text=text, color="white", attrs=["bold"])


def print_warn(text):
    cprint(text=text, color="red", attrs=["bold"])


def print_error(text):
    cprint(text=text, color="red", attrs=["bold"])


def print_title(text):
    cprint(text=text, color="green", attrs=[])


def color_hi(text):
    return colored(text=text, color="white", attrs=["bold"])


def color_warn(text):
    return colored(text=text, color="red", attrs=["bold"])


def color_ok(text):
    return colored(text=text, color="green", attrs=["bold"])



import setuptools


def get_readme():
    with open('README.rst') as f:
        return f.read()


setuptools.setup(
    # the first three fields are a must according to the documentation
    name="pytconf",
    version="0.0.76",
    packages=[
        'pytconf',
    ],
    # from here all is optional
    description="pytconf handles configuration and command line parsing",
    long_description=get_readme(),
    long_description_content_type="text/x-rst",
    author="Mark Veltzer",
    author_email="mark.veltzer@gmail.com",
    maintainer="Mark Veltzer",
    maintainer_email="mark.veltzer@gmail.com",
    keywords=[
        'click',
        'argparse',
        'optparse',
        'command-line-parser',
        'configuration',
    ],
    url="https://veltzer.github.io/pytconf",
    download_url="https://github.com/veltzer/pytconf",
    license="MIT",
    platforms=[
        'python3',
    ],
    install_requires=[
        'termcolor',
        'yattag',
        'pyfakeuse',
    ],
    extras_require={
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
    ],
    data_files=[
    ],
    entry_points={"console_scripts": [
    ]},
    python_requires=">=3.7",
)



from typing import List, Generator


class ErrorsCollector:
    """
    This collector support two types of error. Regular errors and special errors
    Regular errors are the ones to be printed and carry the error_type = false.
    """
    def __init__(self):
        self._errors: List[str] = []
        self.do_help: bool = False
        self.show_errors: bool = True
        self.force_show_errors: bool = False

    def add_error(self, msg: str) -> None:
        self._errors.append(msg)

    def have_errors(self) -> bool:
        return len(self._errors) > 0

    def set_do_help(self) -> None:
        self.do_help = True

    def get_show_errors(self):
        return self.show_errors or self.force_show_errors

    def get_do_help(self) -> bool:
        return self.do_help

    def unset_show_errors(self) -> None:
        self.show_errors = False

    def set_force_show_errors(self) -> None:
        self.force_show_errors = True

    def yield_errors(self) -> Generator[str, None, None]:
        for error in self._errors:
            yield error



from typing import Union


def get_first_line(o, default_val: str) -> Union[str, None]:
    """
    Get first line for a pydoc string
    :param o: object which is documented (class or function)
    :param default_val: value to return if there is no documentation
    :return: the first line which is not whitespace
    """
    doc: Union[str, None] = o.__doc__
    if doc is None:
        return default_val
    lines = doc.split("\n")
    for line in lines:
        if line == "" or line.isspace():
            continue
        return line.strip()
    return default_val



import logging
from yattag import Doc


from pytconf.data import LOGGER_NAME


def get_logger():
    return logging.getLogger(LOGGER_NAME)


def noun(name: str, num: int) -> str:
    """
    This function returns a noun in it's right for a specific quantity
    :param name:
    :param num:
    :return:
    """
    if num == 0 or num > 1:
        return name + "s"
    return name


class HtmlGen:
    def __init__(self):
        document, tag, text, line = Doc().ttl()
        self.document = document
        self.tag = tag
        self.text = text
        self.line = line



""" version which can be consumed from within the module """
VERSION_STR = "0.0.76"
DESCRIPTION = "pytconf handles configuration and command line parsing"
APP_NAME = "pytconf"
LOGGER_NAME = "pytconf"



import json
import os
import sys
from collections import defaultdict
from typing import Union, List, Any, Callable, Type, Dict, Set

from pytconf.color_utils import (
    print_highlight,
    color_hi,
    print_title,
    color_ok,
    color_warn,
    print_error,
)
from pytconf.errors_collector import ErrorsCollector
from pytconf.param_collector import the_collector
from pytconf.pydoc import get_first_line
from pytconf.registry import the_registry
from pytconf.utils import get_logger, noun, HtmlGen

DEFAULT_FUNCTION_GROUP_NAME = "default"
DEFAULT_FUNCTION_GROUP_DESCRIPTION = "default command group"
DEFAULT_FUNCTION_GROUP_SHOW_META = False
DEFAULT_FUNCTION_GROUP_SHOW = True

SPECIAL_FUNCTION_GROUP_NAME = "special"
SPECIAL_FUNCTION_GROUP_DESCRIPTION = "special command group"
SPECIAL_FUNCTION_GROUP_SHOW_META = False
SPECIAL_FUNCTION_GROUP_SHOW = False

DEFAULT_FUNCTION_DESCRIPTION = "No function description available"


class MetaConfig(type):
    """
    Meta class for all configs
    """
    def __new__(mcs, name, bases, namespace):
        ret = super().__new__(mcs, name, bases, namespace)
        i = 0
        for k, v in namespace.items():
            if not k.startswith("__") and not isinstance(v, classmethod):
                the_registry.register(ret, k, the_collector.get_item(i))
                i += 1
        the_collector.clear()
        return ret


class Config(metaclass=MetaConfig):
    """
    Base class for all configs
    """


def is_help(string: str) -> bool:
    return string.lower()[:4] == "help"


class PytconfConf:
    def __init__(self):
        self.main_function: Union[Callable, None] = None
        self.main_description: str = "No application description"
        self.function_name_to_configs: Dict[str, List[Type[Config]]] = {}
        self.function_name_to_suggest_configs: Dict[str, List[Type[Config]]] = {}
        self.function_name_to_callable: Dict[str, Callable] = {}
        self.allow_free_args: Dict[str, bool] = {}
        self.min_free_args: Dict[str, Union[int, None]] = {}
        self.max_free_args: Dict[str, Union[int, None]] = {}
        self.description: Dict[str, str] = {}

        self.function_group_names: Dict[str, Set[str]] = defaultdict(set)
        self.function_group_list: List[str] = []
        self.function_group_descriptions: Dict[str, str] = {}
        self.function_group_show_meta: Dict[str, bool] = {}
        self.function_group_show: Dict[str, bool] = {}

        self.free_args: List[str] = []
        self.app_name: str = "No application name"
        self.version: str = "No version"
        self.default_function = None

    def register_main(
        self,
        main_function: Callable,
        main_description: str,
        app_name: str,
        version: str,
    ):
        self.main_function = main_function
        self.main_description = main_description
        self.app_name = app_name
        self.version = version

    def register_function(
        self,
        function: Callable,
        description: str,
        name: str,
        configs: List[Type[Config]] = (),
        suggest_configs: List[Type[Config]] = (),
        group: str = DEFAULT_FUNCTION_GROUP_NAME,
        allow_free_args: bool = False,
        min_free_args: Union[int, None] = None,
        max_free_args: Union[int, None] = None,
    ):
        self.description[name] = description
        self.function_name_to_callable[name] = function
        self.function_name_to_configs[name] = configs
        self.function_name_to_suggest_configs[name] = suggest_configs
        self.function_group_names[group].add(name)
        self.allow_free_args[name] = allow_free_args
        self.min_free_args[name] = min_free_args
        self.max_free_args[name] = max_free_args

    @classmethod
    def print_errors(cls, errors: ErrorsCollector) -> None:
        for error in errors.yield_errors():
            print_error(error)

    def show_help(self) -> None:
        print(f"Usage: {color_hi(self.app_name)} COMMAND [ARGS]...")
        space = " " * 2
        print_highlight(f"\n{space}{self.main_description}\n")
        for function_group in self.function_group_list:
            show = self.function_group_show[function_group]
            if not show:
                continue
            show_meta = self.function_group_show_meta[function_group]
            if show_meta:
                description = self.function_group_descriptions[function_group]
                print(f"{space}{function_group}: {description}")
                cmd_space = space + " " * 2
            else:
                cmd_space = space
            for name in sorted(self.function_group_names[function_group]):
                function_doc = self.description[name]
                print(f"{cmd_space}{color_hi(name)}: {function_doc}")
            print()

    def show_help_for_function(
        self,
        function_name: str,
        show_help_full: bool = False,
        show_help_suggest: bool = False,
    ) -> None:
        print(f"Usage: {self.app_name} {function_name} [OPTIONS] [ARGS]...")
        function_doc = self.description[function_name]
        print_highlight(f"\n  {function_doc}")
        print("\nOptions:\n")
        for config in self.function_name_to_configs[function_name]:
            self.show_help_for_config(config)
        if show_help_suggest:
            for config in self.function_name_to_suggest_configs[function_name]:
                self.show_help_for_config(config)
        if show_help_full:
            for config in the_registry.yield_configs():
                self.show_help_for_config(config)

    @classmethod
    def show_help_for_config(cls, config):
        if config == Config:
            return
        doc = get_first_line(config, "Undocumented parameter set")
        print_title(f"  {doc}")
        for name, param in the_registry.yield_name_data_for_config(config):
            if param.required:
                default = color_warn("MANDATORY")
            else:
                default = color_ok(param.t2s(param.default))
            print(f"    {color_hi(name)} [{param.get_type_name()}]: {param.help_string} [{default}]")
            more_help = param.more_help()
            if more_help is not None:
                print(f"      {more_help}")
        print()

    def process_flags(
        self, function_selected: str, flags: Dict[str, str], errors: ErrorsCollector,
    ) -> None:
        """
        Parse the args and fill the global data
        :param function_selected:
        :param flags:
        :param errors:
        """

        # set the flags into the "default" field and collect unknown flags
        unknown_flags = []
        for flag_raw, value in flags.items():
            if not the_registry.has_name(flag_raw):
                if is_help(flag_raw):
                    errors.set_do_help()
                    errors.unset_show_errors()
                else:
                    unknown_flags.append(flag_raw)
                continue
            config = the_registry.get_config_for_name(flag_raw)
            param = the_registry.get_data_for_name(flag_raw)
            edit = value.startswith("=")
            # noinspection PyBroadException
            # pylint: disable=broad-except
            try:
                if edit:
                    v = param.s2t_generate_from_default(value[1:])
                else:
                    v = param.s2t(value)
                setattr(config, flag_raw, v)
            except Exception:
                errors.add_error(f"could not convert [{flag_raw}]")
        if unknown_flags:
            errors.add_error(f"unknown flags [{','.join(unknown_flags)}]")

        # check for missing parameters
        missing_parameters = []
        configs = self.function_name_to_configs[function_selected]
        for config in configs:
            for name, param in the_registry.yield_name_data_for_config(config):
                if param.required and name not in flags:
                    missing_parameters.append(name)
        if missing_parameters:
            errors.add_error(
                f"missing {noun('parameter', len(missing_parameters))} [{','.join(missing_parameters)}]"
            )

    @staticmethod
    def read_flags_from_config(file_name: str, flags: Dict[str, str]) -> None:
        if os.path.isfile(file_name):
            with open(file_name, "rt") as json_file:
                new_flags: Dict[str, str] = json.load(json_file)
                assert isinstance(flags, dict)
            for k, v in new_flags.items():
                flags[k] = v

    def get_system_config(self):
        return f"/etc/{self.app_name}.json"

    def get_user_config(self):
        return os.path.expanduser(f"~/.config/{self.app_name}.json")

    def register_function_group(
        self,
        name: str,
        description: str,
        show_meta: bool,
        show: bool,
    ):
        self.function_group_list.append(name)
        self.function_group_descriptions[name] = description
        self.function_group_show_meta[name] = show_meta
        self.function_group_show[name] = show

    def register_defaults(self):
        self.register_function_group(
            name=DEFAULT_FUNCTION_GROUP_NAME,
            description=DEFAULT_FUNCTION_GROUP_DESCRIPTION,
            show_meta=DEFAULT_FUNCTION_GROUP_SHOW_META,
            show=DEFAULT_FUNCTION_GROUP_SHOW,
        )
        self.register_function_group(
            name=SPECIAL_FUNCTION_GROUP_NAME,
            description=SPECIAL_FUNCTION_GROUP_DESCRIPTION,
            show_meta=SPECIAL_FUNCTION_GROUP_SHOW_META,
            show=SPECIAL_FUNCTION_GROUP_SHOW,
        )

    def get_function_selected(self, args: List[str], errors) -> Union[str, None]:
        function_selected = None
        if len(args) > 0:
            command = args.pop(0)
            if command in self.function_name_to_callable:
                function_selected = command
            else:
                if is_help(command):
                    errors.set_do_help()
                    errors.unset_show_errors()
                else:
                    errors.add_error(f"unknown command [{command}]")
                    errors.set_force_show_errors()
        return function_selected

    def config_arg_parse_and_launch(
        self,
        args: Union[List[str], None] = None,
        launch=True,
        do_exit=True,
    ) -> None:

        self.register_defaults()
        if args is None:
            args = sys.argv[1:]

        flags: Dict[str, str] = {}
        errors = ErrorsCollector()
        self.free_args = []

        # read config files
        self.read_flags_from_config(file_name=self.get_system_config(), flags=flags)
        self.read_flags_from_config(file_name=self.get_user_config(), flags=flags)

        function_selected = self.get_function_selected(args, errors)

        # now parse the args
        self.parse_args(args, errors, flags)

        # if we have command we can check free args errors
        if function_selected is not None:
            if self.allow_free_args[function_selected]:
                min_args = self.min_free_args[function_selected]
                if min_args is not None:
                    if len(self.free_args) < min_args:
                        errors.add_error(f"too few free args - {min_args} required")
                max_args = self.max_free_args[function_selected]
                if max_args is not None:
                    if len(self.free_args) >= max_args:
                        errors.add_error(f"too many free args - {max_args} required")
            else:
                if len(self.free_args) > 0:
                    if len(self.free_args) == 1 and is_help(self.free_args[0]):
                        errors.set_do_help()
                        errors.unset_show_errors()
                    else:
                        errors.add_error(f"free args are not allowed [{','.join(self.free_args)}]")

        if function_selected is None:
            errors.add_error("no command is selected")
            errors.set_do_help()
            errors.unset_show_errors()
        else:
            self.process_flags(function_selected, flags, errors)

        if errors.have_errors() or errors.get_do_help():
            if errors.get_show_errors():
                self.print_errors(errors)
            if errors.get_do_help():
                if function_selected:
                    self.show_help_for_function(function_name=function_selected)
                else:
                    self.show_help()
            if do_exit:
                sys.exit(1)
            return

        if launch:
            function_to_run = self.function_name_to_callable[function_selected]
            function_to_run()

    def parse_args(self, args, errors, flags):
        free_args_started = False
        while args:
            current = args.pop(0)
            if current.startswith("--") and not free_args_started:
                real = current[2:]
                number_of_equals = real.count("=")
                if number_of_equals == 1:
                    flag_name, flag_value = real.split("=")
                    flags[flag_name] = flag_value
                elif number_of_equals == 0:
                    if args:
                        more = args.pop(0)
                        flags[real] = more
                    else:
                        errors.add_error(
                            f"argument [{real}] needs a follow-up argument"
                        )
                else:
                    errors.add_error(f"can not parse argument [{real}]")
            else:
                self.free_args.append(current)
                free_args_started = True

    def get_html(self) -> str:
        html_gen = HtmlGen()
        main_doc = self.main_description
        html_gen.line("h1", main_doc)
        html_gen.line("h2", "API specifications")
        single_group = len(self.function_group_list) == 1
        with html_gen.tag("ul"):
            for function_group_name in self.function_group_list:
                if not single_group:
                    function_group_description = self.function_group_descriptions[
                        function_group_name
                    ]
                    html_gen.line("li", function_group_name, title="function group name: ")
                    html_gen.line(
                        "li",
                        function_group_description,
                        title="function group description: ",
                    )
                with html_gen.tag("li"):
                    for function_name in sorted(
                        self.function_group_names[function_group_name]
                    ):
                        self.get_html_for_function(function_name, html_gen)
        # return html_gen.document.getvalue()
        return html_gen.document

    def get_html_for_function(self, function_name, html_gen):
        with html_gen.tag("ul"):
            f = self.function_name_to_callable[function_name]
            function_doc = get_first_line(f, "no description for this function")
            html_gen.line("li", function_name, title="function name: ")
            html_gen.line("li", function_doc, title="function description: ")
            with html_gen.tag("li"):
                with html_gen.tag("ul"):
                    for config in self.function_name_to_configs[function_name]:
                        with html_gen.tag("li"):
                            self.get_html_for_config(config, html_gen)

    @classmethod
    def get_html_for_config(cls, config, html_gen):
        if config == Config:
            return
        doc = get_first_line(config, "undocumented config")
        html_gen.line("h3", doc, title="config: ")
        with html_gen.tag("table"):
            for name, param in the_registry.yield_name_data_for_config(config):
                with html_gen.tag("tr"):
                    if param.required:
                        default = "MANDATORY"
                    else:
                        default = param.t2s(param.default)
                    if param.more_help() is None:
                        more_help = "No more help is documented"
                    else:
                        more_help = param.more_help()
                    html_gen.line("td", name)
                    html_gen.line("td", param.help_string)
                    html_gen.line("td", param.get_type_name())
                    html_gen.line("td", default)
                    html_gen.line("td", more_help)

    @classmethod
    def write_config_file_json(cls, filename: str) -> None:
        values: Dict[str, str] = {}
        for config in the_registry.yield_configs():
            for name, param in the_registry.yield_name_data_for_config(config):
                values[name] = param.t2s(param.default)
        with open(filename, "wt") as f:
            json.dump(values, f, indent=4)

    def write_config_file_json_user(self) -> None:
        self.write_config_file_json(self.get_user_config())

    def write_config_file_json_system(self) -> None:
        self.write_config_file_json(self.get_system_config())


_pytconf = PytconfConf()


def get_pytconf():
    return _pytconf


def config_arg_parse_and_launch(
    args=None,
    launch=True,
    do_exit=True,
) -> None:
    """
    This is the real API
    """
    get_pytconf().config_arg_parse_and_launch(
        args=args,
        launch=launch,
        do_exit=do_exit,
    )


def get_free_args() -> List[str]:
    return get_pytconf().free_args


def register_function_group(
    name: str,
    description: str,
    show_meta: bool,
    show: bool,
) -> None:
    get_pytconf().register_function_group(
        name=name,
        description=description,
        show_meta=show_meta,
        show=show,
    )


def register_main(
    main_description: str,
    app_name: str,
    version: str,
) -> Callable[[Any], Any]:
    def identity(main_function):
        get_pytconf().register_main(
            main_function=main_function,
            main_description=main_description,
            app_name=app_name,
            version=version,
        )
        return main_function

    return identity


def register_endpoint(
    description: str,
    name: Union[str, None] = None,
    configs: List[Type[Config]] = (),
    suggest_configs: List[Type[Config]] = (),
    group: str = DEFAULT_FUNCTION_GROUP_NAME,
    allow_free_args: bool = False,
    min_free_args: Union[int, None] = None,
    max_free_args: Union[int, None] = None,
) -> Callable[[Any], Any]:
    logger = get_logger()
    logger.debug("registering endpoint")

    def identity(function):
        if name is None:
            function_name = function.__name__
        else:
            function_name = name
        register_function(
            function=function,
            description=description,
            name=function_name,
            configs=configs,
            suggest_configs=suggest_configs,
            group=group,
            allow_free_args=allow_free_args,
            min_free_args=min_free_args,
            max_free_args=max_free_args,
        )
        return function

    return identity


def register_function(
    function: Callable,
    description: str,
    name: str,
    configs: List[Type[Config]] = (),
    suggest_configs: List[Type[Config]] = (),
    group: str = DEFAULT_FUNCTION_GROUP_NAME,
    allow_free_args: bool = False,
    min_free_args: Union[int, None] = None,
    max_free_args: Union[int, None] = None,
) -> None:
    get_pytconf().register_function(
        function=function,
        description=description,
        name=name,
        configs=configs,
        suggest_configs=suggest_configs,
        group=group,
        allow_free_args=allow_free_args,
        min_free_args=min_free_args,
        max_free_args=max_free_args,
    )


def write_config_file_json_user() -> None:
    filename = get_pytconf().get_user_config()
    if not os.path.isfile(filename):
        get_pytconf().write_config_file_json_user()


def rm_config_file_json_user() -> None:
    filename = get_pytconf().get_user_config()
    if os.path.isfile(filename):
        os.unlink(filename)


def write_config_file_json_system() -> None:
    filename = get_pytconf().get_system_config()
    if not os.path.isfile(filename):
        get_pytconf().write_config_file_json_system()


def rm_config_file_json_system() -> None:
    filename = get_pytconf().get_system_config()
    if os.path.isfile(filename):
        os.unlink(filename)
